from mailers import Email, Plugin, SentMessages, TemplatedEmail
from mailers.message import Recipients

from kupala.application import get_current_application

try:
    import toronado
except ImportError:
    toronado = None


class CSSInlinerPlugin(Plugin):
    def __init__(self) -> None:
        assert toronado, 'CSSInlinerPlugin depends on "toronado" package.'

    def process_email(self, email: Email) -> Email:
        if email.html:
            email.html = toronado.from_string(email.html).decode()
        return email


async def send_mail(
    message: Email = None,
    *,
    to: Recipients = None,
    subject: str = None,
    from_address: Recipients = None,
    cc: Recipients = None,
    bcc: Recipients = None,
    reply_to: Recipients = None,
    headers: dict = None,
    sender: str = None,
    html: str = None,
    text: str = None,
    return_path: str = None,
    text_charset: str = 'utf-8',
    html_charset: str = 'utf-8',
    boundary: str = None,
    message_id: str = None,
) -> SentMessages:
    message = message or Email(
        to=to,
        subject=subject,
        from_address=from_address,
        cc=cc,
        bcc=bcc,
        reply_to=reply_to,
        headers=headers,
        sender=sender,
        text=text,
        html=html,
        return_path=return_path,
        text_charset=text_charset,
        html_charset=html_charset,
        boundary=boundary,
        message_id=message_id,
    )

    app = get_current_application()
    return await app.mail.send(message)


async def send_templated_mail(
    message: Email = None,
    *,
    to: Recipients = None,
    subject: str = None,
    from_address: Recipients = None,
    cc: Recipients = None,
    bcc: Recipients = None,
    reply_to: Recipients = None,
    headers: dict = None,
    sender: str = None,
    html_template: str = None,
    text_template: str = None,
    context: dict = None,
    return_path: str = None,
    text_charset: str = 'utf-8',
    html_charset: str = 'utf-8',
    boundary: str = None,
    message_id: str = None,
) -> SentMessages:
    message = message or TemplatedEmail(
        to=to,
        subject=subject,
        from_address=from_address,
        cc=cc,
        bcc=bcc,
        reply_to=reply_to,
        headers=headers,
        sender=sender,
        html_template=html_template,
        text_template=text_template,
        context=context,
        return_path=return_path,
        text_charset=text_charset,
        html_charset=html_charset,
        boundary=boundary,
        message_id=message_id,
    )

    app = get_current_application()
    return await app.mail.send(message)



from __future__ import annotations

import click
import itsdangerous.signer
import jinja2
import jinja2.ext
import logging
import os
import time
import typing
from click.testing import CliRunner
from email.message import Message
from functools import cached_property
from imia import BaseAuthenticator, LoginManager, UserProvider, UserToken
from mailers import Email, Encrypter, Mailer, Plugin, SentMessages, Signer, create_transport_from_url

from kupala.console.application import ConsoleApplication
from kupala.contracts import PasswordHasher, TemplateRenderer
from kupala.di import make_app_injectable, make_request_injectable
from kupala.storages.storages import LocalStorage, S3Storage, Storage
from kupala.templating import JinjaRenderer
from kupala.utils import import_string, resolve_path

if typing.TYPE_CHECKING:  # pragma: nocover
    from kupala.application import Kupala


class Extension:
    def initialize(self, app: Kupala) -> None:
        pass


_PasswordHasherType = typing.Literal['pbkdf2_sha256', 'pbkdf2_sha512', 'argon2', 'bcrypt', 'des_crypt'] | PasswordHasher


class PasswordsExtension(Extension):
    def __init__(self, backend: _PasswordHasherType = 'pbkdf2_sha256') -> None:
        self._manager = self._create(backend)

    def hash(self, plain_password: str) -> str:
        return self._manager.hash(plain_password)

    def verify(self, plain: str, hashed: str) -> bool:
        return self._manager.verify(plain, hashed)

    def use(self, backend: _PasswordHasherType) -> None:
        self._manager = self._create(backend)

    def _create(self, backend: _PasswordHasherType) -> PasswordHasher:
        if isinstance(backend, str):
            imports = {
                'pbkdf2_sha256': 'passlib.handlers.pbkdf2:pbkdf2_sha256',
                'pbkdf2_sha512': 'passlib.handlers.pbkdf2:pbkdf2_sha512',
                'argon2': 'passlib.handlers.argon2:argon2',
                'bcrypt': 'passlib.handlers.bcrypt:bcrypt',
                'des_crypt': 'passlib.handlers.des_crypt:des_crypt',
            }
            backend = typing.cast(PasswordHasher, import_string(imports[backend]))
        return backend

    def initialize(self, app: Kupala) -> None:
        app.di.prefer_for(PasswordHasher, lambda app: self._manager)


class RendererExtension(Extension):
    def __init__(self, renderer: TemplateRenderer | None = None) -> None:
        self._template_renderer = renderer

    def use(self, renderer: TemplateRenderer) -> None:
        self._template_renderer = renderer

    def render(self, template_name: str, context: dict[str, typing.Any] = None) -> str:
        assert self._template_renderer, 'Template rendering is not configured.'
        return self._template_renderer.render(template_name, context)


class MailExtension(Extension):
    def __init__(self, mailers: dict[str, Mailer] = None, default: str = 'default') -> None:
        self.default = default
        self._mailers: dict[str, Mailer] = mailers or {}

    def get(self, name: str) -> Mailer:
        if name not in self._mailers:
            raise KeyError(f'No mailer named "{name}" defined.')
        return self._mailers[name]

    def use(
        self,
        url: str,
        *,
        from_address: str = 'no-reply@example.com',
        from_name: str = 'Example',
        signer: Signer = None,
        encrypter: Encrypter = None,
        plugins: list[Plugin] = None,
        name: str = 'default',
    ) -> None:
        if from_name:
            from_address = f'{from_name} <{from_address}>'
        transport = create_transport_from_url(url)
        self.add(
            name,
            Mailer(
                transport,
                from_address=from_address,
                plugins=plugins,
                signer=signer,
                encrypter=encrypter,
            ),
        )

    def add(self, name: str, mailer: Mailer) -> None:
        assert name not in self._mailers, f'"{name}" already exists.'
        self._mailers[name] = mailer

    def get_default(self) -> Mailer:
        return self.get(self.default)

    def initialize(self, app: Kupala) -> None:
        make_app_injectable(Mailer, lambda app: self.get_default())

    async def send(self, message: Email | Message, mailer: str = None) -> SentMessages:
        mailer_name = mailer or self.default
        return await self.get(mailer_name).send(message)


class AuthenticationExtension(Extension):
    def __init__(self, app: Kupala) -> None:
        self.user_model: typing.Type[typing.Any] | None = None
        self.user_provider: UserProvider | None = None
        self.authenticators: list[BaseAuthenticator] = []
        self.password_verifier = app.passwords
        self.secret_key = app.secret_key

    @property
    def login_manager(self) -> LoginManager:
        assert self.user_provider, 'Authentication component requires user provider to be configured.'
        return LoginManager(
            user_provider=self.user_provider,
            password_verifier=self.password_verifier,
            secret_key=self.secret_key,
        )

    def configure(
        self,
        user_model: typing.Type[typing.Any] = None,
        user_provider: UserProvider | None = None,
        authenticators: list[BaseAuthenticator] | None = None,
    ) -> None:
        """Configure multiple options at once."""
        if user_model:
            self.user_model = user_model
        if user_provider:
            self.user_provider = user_provider
        if authenticators:
            self.authenticators = authenticators

    def initialize(self, app: Kupala) -> None:
        make_app_injectable(LoginManager, lambda app: self.login_manager)
        make_request_injectable(UserToken, lambda request: request.auth)


class StoragesExtension(Extension):
    def __init__(self, storages: dict[str, Storage] = None, default: str = 'default') -> None:
        self.default = default
        self._storages: dict[str, Storage] = storages or {}

    def get(self, name: str) -> Storage:
        if name not in self._storages:
            raise KeyError(f'No storage named "{name}" defined.')
        return self._storages[name]

    def get_default(self) -> Storage:
        if self.default in self._storages:
            return self._storages[self.default]
        if len(self._storages) == 1:
            return list(self._storages.values())[0]
        raise KeyError('No default storage configured.')

    def add(self, name: str, storage: Storage) -> None:
        self._storages[name] = storage

    def add_local(self, name: str, base_dir: str | os.PathLike) -> None:
        self.add(name, LocalStorage(base_dir))

    def add_s3(
        self,
        name: str,
        bucket: str,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        region_name: str = None,
        profile_name: str = None,
        endpoint_url: str = None,
        link_ttl: int = 300,
    ) -> None:
        self.add(
            name,
            S3Storage(
                bucket=bucket,
                aws_secret_access_key=aws_secret_access_key,
                aws_access_key_id=aws_access_key_id,
                region_name=region_name,
                profile_name=profile_name,
                endpoint_url=endpoint_url,
                link_ttl=link_ttl,
            ),
        )

    def initialize(self, app: Kupala) -> None:
        make_app_injectable(Storage, lambda app: self.get_default())


class JinjaExtension(Extension):
    def __init__(
        self,
        template_dirs: str | list[str] | None = None,
        tests: dict[str, typing.Callable] = None,
        filters: dict[str, typing.Callable] = None,
        globals: dict[str, typing.Any] = None,
        policies: dict[str, typing.Any] = None,
        extensions: list[str | typing.Type[jinja2.ext.Extension]] = None,
        env: jinja2.Environment = None,
        loader: jinja2.BaseLoader = None,
    ) -> None:
        self.template_dirs = [template_dirs] if isinstance(template_dirs, str) else template_dirs or []
        self.tests = tests or {}
        self.filters = filters or {}
        self.globals = globals or {}
        self.policies = policies or {}
        self.extensions = extensions or []
        self._env = env
        self._loader = loader

    @property
    def loader(self) -> jinja2.BaseLoader:
        """Create and return instance of jinja2.BaseLoader.
        If no `loader` constructor argument passed then jinja2.FileSystemLoader will be used."""
        if self._loader is None:
            self._loader = jinja2.FileSystemLoader(
                searchpath=[resolve_path(directory) for directory in self.template_dirs]
            )
        return self._loader

    @property
    def env(self) -> jinja2.Environment:
        """Create and return jinja2.Environment instance."""
        if self._env is None:
            env = jinja2.Environment(loader=self.loader, extensions=self.extensions)
            env.globals.update(self.globals)
            env.filters.update(self.filters)
            env.policies.update(self.policies)
            env.tests.update(self.tests)

            if "json.dumps_kwargs" not in env.policies:
                env.policies["json.dumps_kwargs"] = {"ensure_ascii": False, "sort_keys": True}
            self._env = env
        return self._env

    @cached_property
    def renderer(self) -> JinjaRenderer:
        """Create and return Jinja template renderer."""
        return JinjaRenderer(self.env)

    def use_loader(self, loader: jinja2.BaseLoader) -> None:
        """Use custom template loader."""
        self._loader = loader

    def use_env(self, env: jinja2.Environment) -> None:
        """Use preconfigured loader."""
        self._env = env

    def add_template_dirs(self, directories: list[str | os.PathLike] | str | os.PathLike) -> None:
        """Add additional directories for template search."""
        if isinstance(directories, (str, os.PathLike)):
            directories = [directories]

        if isinstance(self.env.loader, jinja2.FileSystemLoader):
            self.env.loader.searchpath = [resolve_path(str(directory)) for directory in directories]
        else:  # pragma: nocover
            logging.warning('Hot directory update supported only for jinja2.FileSystemLoader loader.')

    def add_globals(self, globals: dict[str, typing.Any]) -> None:
        """Add global variables to jinja environment."""
        self.env.globals.update(globals)

    def add_filters(self, filters: dict[str, typing.Callable]) -> None:
        """Add filters to jinja environment."""
        self.env.filters.update(filters)

    def add_policies(self, policies: dict[str, typing.Any]) -> None:
        """Add policies to jinja environment."""
        self.env.policies.update(policies)

    def add_tests(self, tests: dict[str, typing.Any]) -> None:
        """Add tests to jinja environment."""
        self.env.tests.update(tests)

    def add_extensions(self, *extension: str | typing.Type[jinja2.ext.Extension]) -> None:
        """Add extension to jinja environment."""
        for _extension in extension:
            self.env.add_extension(_extension)

    def configure(
        self,
        template_dirs: list[str | os.PathLike] = None,
        globals: dict[str, typing.Any] = None,
        filters: dict[str, typing.Callable] = None,
        policies: dict[str, typing.Any] = None,
        tests: dict[str, typing.Any] = None,
        extensions: list[str | typing.Type[jinja2.ext.Extension]] = None,
    ) -> None:
        """Configure multiple options at once."""
        if template_dirs:
            self.add_template_dirs(template_dirs)
        if globals:
            self.add_globals(globals)
        if filters:
            self.add_filters(filters)
        if policies:
            self.add_policies(policies)
        if tests:
            self.add_tests(tests)
        if extensions:
            self.add_extensions(*extensions)


class SignerExtension(Extension):
    def __init__(self, secret_key: str) -> None:
        self.secret_key = secret_key

    def initialize(self, app: Kupala) -> None:
        make_app_injectable(itsdangerous.signer.Signer, lambda app: self.signer)
        make_app_injectable(itsdangerous.timed.TimestampSigner, lambda app: self.timestamp_signer)

    @cached_property
    def signer(self) -> itsdangerous.signer.Signer:
        return itsdangerous.signer.Signer(self.secret_key)

    @cached_property
    def timestamp_signer(self) -> itsdangerous.timed.TimestampSigner:
        return itsdangerous.timed.TimestampSigner(self.secret_key)

    def sign(self, value: str | bytes) -> bytes:
        """Sign value."""
        return self.signer.sign(value)

    def unsign(self, signed_value: str | bytes) -> bytes:
        """Unsign value. Raises itsdangerous.BadSignature exception."""
        return self.signer.unsign(signed_value)

    def safe_unsign(self, signed_value: str | bytes) -> tuple[bool, typing.Optional[bytes]]:
        """Unsign value. Will not raise itsdangerous.BadSignature.
        Returns two-tuple: operation status and unsigned value."""
        try:
            return True, self.unsign(signed_value)
        except itsdangerous.BadSignature:
            return False, None

    def timed_sign(self, value: str | bytes) -> bytes:
        """Sign value. The signature will be valid for a specific time period."""
        return self.timestamp_signer.sign(value)

    def timed_unsign(self, signed_value: str | bytes, max_age: int) -> bytes:
        """Unsign value. Will raise itsdangerous.BadSignature or itsdangerous.BadTimeSignature exception
        if signed value cannot be decoded or expired.."""
        return self.timestamp_signer.unsign(signed_value, max_age)

    def safe_timed_unsign(self, signed_value: str | bytes, max_age: int) -> tuple[bool, typing.Optional[bytes]]:
        """Unsign value. Will not raise itsdangerous.BadTimeSignature.
        Returns two-tuple: operation status and unsigned value."""
        try:
            return True, self.timed_unsign(signed_value, max_age)
        except itsdangerous.BadSignature:
            return False, None


class ConsoleExtension:
    def __init__(self, app: Kupala, commands: list[click.Command] = None) -> None:
        self.app = app
        self.commands = commands or []

    @property
    def test_runner(self) -> CliRunner:
        return CliRunner()

    def add(self, *command: click.Command) -> None:
        self.commands.extend(command)

    def run(self) -> int:
        app = ConsoleApplication(self.app, self.commands)
        return app.run()

    def __iter__(self) -> typing.Iterator[click.Command]:
        return iter(self.commands)


class StaticFilesExtension(Extension):
    """Configures static files endpoint."""

    def __init__(self, app: Kupala) -> None:
        self.app = app
        self.start_time = time.time()

    def serve_from_directory(
        self,
        directory: str | os.PathLike = 'statics',
        url_path: str = '/static',
        url_prefix: str = '',
        storage_name: str = 'static',
        path_name: str = 'static',
        random_suffix: bool = False,
    ) -> None:
        assert self.app, 'Unbound StaticFiles instance.'
        self.url_prefix = url_prefix
        self.path_name = path_name
        self.random_suffix = random_suffix
        self.app.storages.add_local(storage_name, directory)
        self.app.routes.files(url_path, storage=storage_name, name=path_name, inline=False)
        self.app.jinja.add_globals({"static": self.static_url})

    def static_url(self, path: str | os.PathLike) -> str:
        router = self.app.get_asgi_app().router
        url = str(router.url_path_for(self.path_name, path=str(path)))
        if not url.startswith('http') and self.url_prefix:
            url = self.url_prefix.rstrip('/') + url

        if self.random_suffix:
            url += '?' + str(self.start_time)
        return url


class URLExtension(Extension):
    def __init__(self, app: Kupala) -> None:
        self._app = app

    def url_for(self, path_name: str, **params: str) -> str:
        return self._app.get_asgi_app().router.url_path_for(path_name, **params)



from __future__ import annotations

import typing
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.middleware import MiddlewareStack
from kupala.middleware.errors import ServerErrorMiddleware
from kupala.middleware.exception import ErrorHandler, ExceptionMiddleware
from kupala.requests import Request
from kupala.responses import Response
from kupala.routing import Router, Routes


class ASGIHandler:
    def __init__(
        self,
        *,
        debug: bool,
        routes: Routes,
        middleware: MiddlewareStack,
        error_handlers: dict[typing.Type[Exception] | int, typing.Optional[ErrorHandler]],
        exception_handler: typing.Callable[[Request, Exception], Response] | None,
    ) -> None:
        self.debug = debug
        self.middleware = middleware
        self.routes = Routes(routes or [])
        self.error_handlers = error_handlers
        self.exception_handler = exception_handler

        middleware.use(ExceptionMiddleware, handlers=error_handlers)
        middleware.top(ServerErrorMiddleware, debug=debug, handler=self.exception_handler)
        self.router = Router(routes=routes)
        self.asgi_app: ASGIApp = self.router
        for mw in reversed(middleware):
            self.asgi_app = mw.wrap(self.asgi_app)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope['type'] in {'http', 'websocket', 'lifespan'}
        await self.asgi_app(scope, receive, send)



import dataclasses

import collections.abc
import datetime
import decimal
import enum
import functools
import json
import typing
import uuid
from babel.support import LazyProxy

JSONEncoder = json.JSONEncoder

_type_to_encoder: dict[type, typing.Callable] = {
    uuid.UUID: str,
    datetime.timedelta: str,
    datetime.datetime: lambda x: x.isoformat(),
    datetime.date: lambda x: x.isoformat(),
    datetime.time: lambda x: x.isoformat(),
    set: list,
    frozenset: list,
    collections.abc.KeysView: list,
    collections.abc.ValuesView: list,
    decimal.Decimal: str,
    LazyProxy: str,
    enum.Enum: lambda x: x.value,
}


def json_default(o: typing.Any) -> typing.Any:
    """Usage: json.dumps(data, default=json_default)"""
    if hasattr(o, "to_json"):
        return o.to_json()

    if hasattr(o, "__json__"):
        return o.__json__()

    for type_class, encoder in _type_to_encoder.items():
        if isinstance(o, type_class):
            return encoder(o)

    if dataclasses.is_dataclass(o):
        return dataclasses.asdict(o)

    return JSONEncoder().default(o)


def walk(value: typing.Any, callback: typing.Callable) -> typing.Any:
    """Visit each item and apply"""
    if isinstance(value, dict):
        for key, value_ in value.items():
            value[key] = walk(value_, callback)

    if isinstance(value, (list, tuple, set)):
        return list(walk(x, callback) for x in value)

    if dataclasses.is_dataclass(value):
        value = dataclasses.asdict(value)
        return walk(value, callback)

    try:
        return callback(value)
    except TypeError:
        return value


def jsonify(value: typing.Any) -> typing.Any:
    """Visit each item and convert it into JSON serializable value."""
    return walk(value, functools.partial(json_default))


JSONData = typing.Any  # https://github.com/python/typing/issues/182


def dump(
    value: JSONData,
    fp: typing.IO[str],
    default: typing.Callable[[typing.Any], typing.Any] = json_default,
    **kwargs: typing.Any,
) -> None:
    json.dump(value, fp, default=default, **kwargs)


def dumps(value: JSONData, **kwargs: typing.Any) -> str:
    if "default" not in kwargs and "cls" not in kwargs:
        kwargs["default"] = json_default
    return json.dumps(value, **kwargs)


load = json.load
loads = json.loads



from __future__ import annotations

from kupala.application import get_current_application


def make_password(raw_password: str) -> str:
    return get_current_application().passwords.hash(raw_password)


def check_password(raw_password: str, hashed_password: str) -> bool:
    return get_current_application().passwords.verify(raw_password, hashed_password)



from dataclasses import dataclass, field

import inspect
import typing
import typing as t
from contextlib import AsyncExitStack, ExitStack, asynccontextmanager, contextmanager
from starlette.concurrency import run_in_threadpool
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.di import InjectionError, get_request_injection_factory
from kupala.middleware import Middleware
from kupala.requests import Request
from kupala.responses import EmptyResponse, HTMLResponse, JSONResponse, PlainTextResponse, Response
from kupala.templating import TemplateResponse
from kupala.utils import run_async


def _callable_name(injection: typing.Any) -> str:
    class_name = injection.__name__ if inspect.isclass(injection) else getattr(injection, '__name__', repr(injection))
    module_name = getattr(injection, '__module__', '')
    return f'{module_name}.{class_name}{inspect.signature(injection)}'


@dataclass
class ActionConfig:
    """Keeps endpoint configuration."""

    methods: t.List[str] = field(default_factory=list)
    template: t.Optional[str] = None
    middleware: t.Optional[t.Sequence[Middleware]] = None


def action_config(
    methods: t.List[str] = None,
    template: str = '',
    middleware: t.Sequence[Middleware] = None,
) -> t.Callable:
    """Use this decorator to configure endpoint parameters."""
    allowed_methods = methods or ['GET', 'HEAD']

    def wrapper(fn: t.Callable) -> t.Callable:
        setattr(
            fn, '__action_config__', ActionConfig(methods=allowed_methods, template=template, middleware=middleware)
        )
        return fn

    return wrapper


def get_action_config(endpoint: t.Callable) -> t.Optional[ActionConfig]:
    """Return endpoint config if defined otherwise return None."""
    return getattr(endpoint, '__action_config__', None)


def detect_request_class(endpoint: t.Callable) -> t.Type[Request]:
    """Detect which request class to use for this endpoint.
    If endpoint does not have `request` argument, or it is not type-hinted
    then default request class returned."""
    args = t.get_type_hints(endpoint)
    return args.get('request', Request)


async def resolve_injections(
    request: Request,
    endpoint: t.Callable,
    sync_stack: ExitStack,
    async_stack: AsyncExitStack,
) -> dict[str, t.Any]:
    """Read endpoint signature and extract injections types.
    These injections will be resolved into actual service instances.
    Dependency injections and path parameters are merged.

    Return value of `from_request` can be a generator. In this case we convert it into context manager
    and add to sync/async exit stack.
    """
    injections = {}

    args = t.get_type_hints(endpoint)
    signature = inspect.signature(endpoint)
    for arg_name, arg_type in args.items():
        if arg_name == 'return':
            continue

        if arg_type == type(request):
            injections[arg_name] = request
            continue

        if arg_name in request.path_params:
            injections[arg_name] = request.path_params[arg_name]
            continue

        if callback := get_request_injection_factory(arg_type):
            if inspect.isgeneratorfunction(callback):
                injections[arg_name] = sync_stack.enter_context(contextmanager(callback)(request))
                continue

            if inspect.isasyncgenfunction(callback):
                injections[arg_name] = await async_stack.enter_async_context(asynccontextmanager(callback)(request))
                continue

            injections[arg_name] = await run_async(callback, request)
            continue

        try:
            injection = request.app.di.make(arg_type)
            if inspect.iscoroutine(injection):
                injection = await injection
            injections[arg_name] = injection
        except InjectionError as ex:
            if arg_name in signature.parameters and signature.parameters[arg_name].default != signature.empty:
                injections[arg_name] = signature.parameters[arg_name].default
            else:
                raise InjectionError(
                    f'Injection "{arg_name}" cannot be processed in {_callable_name(endpoint)}.'
                ) from ex
        else:
            continue

    return injections


def _guess_response_type(
    request: Request,
    action_config: ActionConfig | None,
    content: t.Any,
    status: int,
    headers: dict[str, str],
) -> Response:
    """When endpoint returns non-response instances, then we try to do our best to guess what response type
    we should send to the client."""
    if request.wants_json:
        return JSONResponse(content=content, status_code=status, headers=headers)

    if 'text/html' in request.headers['accept']:
        if action_config and action_config.template:
            return TemplateResponse(
                request,
                template_name=action_config.template,
                context=content,
                status_code=status,
                headers=headers,
            )
        return HTMLResponse(content, status_code=status, headers=headers)

    # default is text/plain, but we can serialize only several types
    return PlainTextResponse(str(content), status_code=status, headers=headers)


def handle_endpoint_result(request: Request, result: t.Any, action_config: ActionConfig | None) -> ASGIApp:
    """Takes return value of endpoint execution and converts it into response instance."""
    if result is None:
        return EmptyResponse()

    if isinstance(result, Response):
        return result

    if inspect.isfunction(result):  # may be request_response function aka ASGI callable
        return result

    status = 200
    headers = {}
    content = result

    if isinstance(result, tuple):
        if len(result) == 3:
            content, status, headers = result
        elif len(result) == 2:
            content, status = result

    return _guess_response_type(request, action_config, content, status, headers)


async def dispatch_endpoint(scope: Scope, receive: Receive, send: Send, endpoint: t.Callable) -> ASGIApp:
    """Call endpoint callable resolving all dependencies. Will return response."""
    request_class = detect_request_class(endpoint)
    request = request_class(scope, receive, send)
    action_config = get_action_config(endpoint)

    with ExitStack() as sync_stack:
        async with AsyncExitStack() as async_stack:
            args = await resolve_injections(request, endpoint, sync_stack, async_stack)
            if inspect.iscoroutinefunction(endpoint):
                response = await endpoint(**args)
            else:
                response = await run_in_threadpool(endpoint, **args)
    return handle_endpoint_result(request, response, action_config)



from __future__ import annotations

import click
import contextvars as cv
import inspect
import logging
import os
import secrets
import traceback
import typing
from contextlib import AsyncExitStack
from starlette.datastructures import State
from starlette.routing import BaseRoute
from starlette.types import Receive, Scope, Send

from kupala.asgi import ASGIHandler
from kupala.config import Config
from kupala.contracts import TemplateRenderer
from kupala.di import Injector
from kupala.exceptions import ShutdownError, StartupError
from kupala.extensions import (
    AuthenticationExtension,
    ConsoleExtension,
    JinjaExtension,
    MailExtension,
    PasswordsExtension,
    RendererExtension,
    SignerExtension,
    StaticFilesExtension,
    StoragesExtension,
    URLExtension,
)
from kupala.middleware import Middleware, MiddlewareStack
from kupala.middleware.exception import ErrorHandler
from kupala.requests import Request
from kupala.responses import Response
from kupala.routing import Routes
from kupala.storages.storages import Storage
from kupala.utils import resolve_path


class Kupala:
    def __init__(
        self,
        *,
        secret_key: str = None,
        debug: bool = False,
        environment: str = 'production',
        middleware: list[Middleware] = None,
        routes: list[BaseRoute] | Routes | None = None,
        request_class: typing.Type[Request] = Request,
        error_handlers: dict[typing.Type[Exception] | int, typing.Optional[ErrorHandler]] = None,
        lifespan_handlers: list[typing.Callable[[Kupala], typing.AsyncContextManager[None]]] = None,
        exception_handler: typing.Callable[[Request, Exception], Response] | None = None,
        template_dir: str | list[str] = 'templates',
        commands: list[click.Command] = None,
        storages: dict[str, Storage] = None,
        renderer: TemplateRenderer = None,
    ) -> None:
        # base configuration
        self.debug = debug
        self.environment = environment
        self.request_class = request_class

        self.secret_key = secret_key
        if self.secret_key is None:
            self.secret_key = secrets.token_hex(128)
            logging.warning(
                'Secret key is not defined. Will generate temporary secret. '
                'This value will be reset after application restart.'
            )

        # ASGI related
        self.error_handlers = error_handlers or {}
        self.lifespan = lifespan_handlers or []
        self.exception_handler = exception_handler
        self.middleware = MiddlewareStack(middleware)
        self.routes = Routes(routes or [])

        # assign core components
        template_dirs = [template_dir] if isinstance(template_dir, (str, os.PathLike)) else template_dir or []
        self.jinja = JinjaExtension(template_dirs=[resolve_path(directory) for directory in template_dirs])
        self.config = Config()
        self.state = State()
        self.passwords = PasswordsExtension()
        self.mail = MailExtension()
        self.auth = AuthenticationExtension(self)
        self.storages = StoragesExtension(storages)
        self.commands = ConsoleExtension(self, commands)
        self.signer = SignerExtension(self.secret_key)
        self.renderer = RendererExtension(renderer or self.jinja.renderer)
        self.staticfiles = StaticFilesExtension(self)
        self.urls = URLExtension(self)
        self.di = Injector(self)

        set_current_application(self)

        # configure app
        self.configure()

        # bind extensions to this app
        self._bootstrap()

        # ASGI app instance
        self._asgi_app: ASGIHandler | None = None

    def _bootstrap(self) -> None:
        for _, extension in inspect.getmembers(self, lambda x: hasattr(x, 'initialize')):
            extension.initialize(self)

    def configure(self) -> None:
        pass

    def create_asgi_app(self) -> ASGIHandler:
        return ASGIHandler(
            debug=self.debug,
            routes=self.routes,
            middleware=self.middleware,
            error_handlers=self.error_handlers,
            exception_handler=self.exception_handler,
        )

    def get_asgi_app(self) -> ASGIHandler:
        """Creates ASGI handler. Subsequent calls will return cached instance."""
        if self._asgi_app is None:
            self._asgi_app = self.create_asgi_app()
        return self._asgi_app

    def render(self, template_name: str, context: dict[str, typing.Any] = None) -> str:
        """Render template."""
        return self.renderer.render(template_name, context or {})

    def url_for(self, name: str, **path_params: str) -> str:
        """Generate URL by name. This method is useful when you want to reverse the URL in non-ASGI mode like CLI.
        Otherwise, prefer using Request.url_for as it generates full URL incl. host and scheme."""
        return self.urls.url_for(name, **path_params)

    def cli(self) -> int:
        """Run console application."""
        return self.commands.run()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope['type'] in {'http', 'websocket', 'lifespan'}

        if scope['type'] == 'lifespan':
            # reset inner ASGI app because underlying routes stack may have been changed between
            # app instantiation and invocation of this method.
            # This may happen when you call app.url_for() or similar methods which instantiate ASGI application
            # to complete
            # TODO: may be this is a bad idea
            self._asgi_app = None
            await self.lifespan_handler(scope, receive, send)
            return

        if self._asgi_app is None:
            self._asgi_app = self.get_asgi_app()

        assert self._asgi_app, 'ASGI application has not been initialized.'

        scope['app'] = self
        scope['state'] = {}
        if scope['type'] == 'http':
            scope['request'] = self.request_class(scope, receive)
        await self._asgi_app(scope, receive, send)

    async def lifespan_handler(self, scope: Scope, receive: Receive, send: Send) -> None:
        started = False
        try:
            await receive()
            async with AsyncExitStack() as stack:
                for hook in self.lifespan:
                    await stack.enter_async_context(hook(self))
                await send({'type': 'lifespan.startup.complete'})
                started = True
                await receive()
        except BaseException as ex:
            text = traceback.format_exc()
            if started:
                await send({'type': 'lifespan.shutdown.failed', 'message': text})
                raise ShutdownError(text) from ex
            else:
                await send({'type': 'lifespan.startup.failed', 'message': text})
                raise StartupError(text) from ex
        else:
            await send({'type': 'lifespan.shutdown.complete'})


_current_app: cv.ContextVar[Kupala] = cv.ContextVar('_current_app')


def set_current_application(app: Kupala) -> None:
    _current_app.set(app)


def get_current_application() -> Kupala:
    return _current_app.get()
